@@ -18,10 +18,14 @@
 from .common import CMakeException
 from .generator import parse_generator_expressions
 from .. import mlog
+from ..mesonlib import version_compare
 
 import typing as T
+from pathlib import Path
 import re
 import os
+import json
+import textwrap
 
 class CMakeTraceLine:
     def __init__(self, file, line, func, args):
@@ -44,6 +48,8 @@ class CMakeTarget:
         self.imported = imported
         self.tline = tline
         self.depends = []
+        self.current_bin_dir = None
+        self.current_src_dir = None
 
     def __repr__(self):
         s = 'CMake TARGET:\n  -- name:      {}\n  -- type:      {}\n  -- imported:  {}\n  -- properties: {{\n{}     }}\n  -- tline: {}'
@@ -52,6 +58,13 @@ class CMakeTarget:
             propSTR += "      '{}': {}\n".format(i, self.properties[i])
         return s.format(self.name, self.type, self.imported, propSTR, self.tline)
 
+    def strip_properties(self) -> None:
+        # Strip the strings in the properties
+        if not self.properties:
+            return
+        for key, val in self.properties.items():
+            self.properties[key] = [x.strip() for x in val]
+
 class CMakeGeneratorTarget(CMakeTarget):
     def __init__(self, name):
         super().__init__(name, 'CUSTOM', {})
@@ -60,24 +73,29 @@ class CMakeGeneratorTarget(CMakeTarget):
         self.working_dir = None  # type: T.Optional[str]
 
 class CMakeTraceParser:
-    def __init__(self, permissive: bool = False):
-        # Dict of CMake variables: '<var_name>': ['list', 'of', 'values']
-        self.vars = {}
+    def __init__(self, cmake_version: str, build_dir: str, permissive: bool = False):
+        self.vars = {}     # type: T.Dict[str, T.List[str]]
+        self.targets = {}  # type: T.Dict[str, CMakeTarget]
 
-        # Dict of CMakeTarget
-        self.targets = {}
+        self.explicit_headers = set()  # type: T.Set[Path]
 
         # T.List of targes that were added with add_custom_command to generate files
         self.custom_targets = []  # type: T.List[CMakeGeneratorTarget]
 
         self.permissive = permissive  # type: bool
+        self.cmake_version = cmake_version  # type: str
+        self.trace_file = 'cmake_trace.txt'
+        self.trace_file_path = Path(build_dir) / self.trace_file
+        self.trace_format = 'json-v1' if version_compare(cmake_version, '>=3.17') else 'human'
 
-    def parse(self, trace: str) -> None:
-        # First parse the trace
-        lexer1 = self._lex_trace(trace)
+        # State for delayed command execution. Delayed command execution is realised
+        # with a custom CMake file that overrides some functions and adds some
+        # introspection information to the trace.
+        self.delayed_commands = []  # type: T.List[str]
+        self.stored_commands = []   # type: T.List[CMakeTraceLine]
 
         # All supported functions
-        functions = {
+        self.functions = {
             'set': self._cmake_set,
             'unset': self._cmake_unset,
             'add_executable': self._cmake_add_executable,
@@ -92,15 +110,63 @@ class CMakeTraceParser:
             'target_link_libraries': self._cmake_target_link_libraries,
             'target_link_options': self._cmake_target_link_options,
             'add_dependencies': self._cmake_add_dependencies,
+
+            # Special functions defined in the preload script.
+            # These functions do nothing in the CMake code, but have special
+            # meaning here in the trace parser.
+            'meson_ps_execute_delayed_calls': self._meson_ps_execute_delayed_calls,
+            'meson_ps_reload_vars': self._meson_ps_reload_vars,
         }
 
+    def trace_args(self) -> T.List[str]:
+        arg_map = {
+            'human': ['--trace', '--trace-expand'],
+            'json-v1': ['--trace-expand', '--trace-format=json-v1'],
+        }
+
+        base_args = ['--no-warn-unused-cli']
+        if not self.requires_stderr():
+            base_args += ['--trace-redirect={}'.format(self.trace_file)]
+
+        return arg_map[self.trace_format] + base_args
+
+    def requires_stderr(self) -> bool:
+        return version_compare(self.cmake_version, '<3.16')
+
+    def parse(self, trace: T.Optional[str] = None) -> None:
+        # First load the trace (if required)
+        if not self.requires_stderr():
+            if not self.trace_file_path.exists and not self.trace_file_path.is_file():
+                raise CMakeException('CMake: Trace file "{}" not found'.format(str(self.trace_file_path)))
+            trace = self.trace_file_path.read_text()
+        if not trace:
+            raise CMakeException('CMake: The CMake trace was not provided or is empty')
+
+        # Second parse the trace
+        lexer1 = None
+        if self.trace_format == 'human':
+            lexer1 = self._lex_trace_human(trace)
+        elif self.trace_format == 'json-v1':
+            lexer1 = self._lex_trace_json(trace)
+        else:
+            raise CMakeException('CMake: Internal error: Invalid trace format {}. Expected [human, json-v1]'.format(self.trace_format))
+
         # Primary pass -- parse everything
         for l in lexer1:
+            # store the function if its execution should be delayed
+            if l.func in self.delayed_commands:
+                self.stored_commands += [l]
+                continue
+
             # "Execute" the CMake function if supported
-            fn = functions.get(l.func, None)
+            fn = self.functions.get(l.func, None)
             if(fn):
                 fn(l)
 
+        # Postprocess
+        for tgt in self.targets.values():
+            tgt.strip_properties()
+
     def get_first_cmake_var_of(self, var_list: T.List[str]) -> T.List[str]:
         # Return the first found CMake variable in list var_list
         for i in var_list:
@@ -116,16 +182,24 @@ class CMakeTraceParser:
 
         return []
 
-    def var_to_bool(self, var):
-        if var not in self.vars:
-            return False
+    def var_to_str(self, var: str) -> T.Optional[str]:
+        if var in self.vars and self.vars[var]:
+            return self.vars[var][0]
 
-        if len(self.vars[var]) < 1:
-            return False
+        return None
 
-        if self.vars[var][0].upper() in ['1', 'ON', 'TRUE']:
-            return True
-        return False
+    def _str_to_bool(self, expr: T.Union[str, T.List[str]]) -> bool:
+        if not expr:
+            return False
+        if isinstance(expr, list):
+            expr_str = expr[0]
+        else:
+            expr_str = expr
+        expr_str = expr_str.upper()
+        return expr_str not in ['0', 'OFF', 'NO', 'FALSE', 'N', 'IGNORE'] and not expr_str.endswith('NOTFOUND')
+
+    def var_to_bool(self, var: str) -> bool:
+        return self._str_to_bool(self.vars.get(var, []))
 
     def _gen_exception(self, function: str, error: str, tline: CMakeTraceLine) -> None:
         # Generate an exception if the parser is not in permissive mode
@@ -240,7 +314,7 @@ class CMakeTraceParser:
 
     def _cmake_add_custom_command(self, tline: CMakeTraceLine, name=None):
         # DOC: https://cmake.org/cmake/help/latest/command/add_custom_command.html
-        args = list(tline.args) # Make a working copy
+        args = self._flatten_args(list(tline.args))  # Commands can be passed as ';' seperated lists
 
         if not args:
             return self._gen_exception('add_custom_command', 'requires at least 1 argument', tline)
@@ -261,7 +335,7 @@ class CMakeTraceParser:
         def handle_command(key: str, target: CMakeGeneratorTarget) -> None:
             if key == 'ARGS':
                 return
-            target.command[-1] += key.split(';')
+            target.command[-1] += [key]
 
         def handle_depends(key: str, target: CMakeGeneratorTarget) -> None:
             target.depends += [key]
@@ -293,6 +367,8 @@ class CMakeTraceParser:
             if fn is not None:
                 fn(i, target)
 
+        target.current_bin_dir = self.var_to_str('MESON_PS_CMAKE_CURRENT_BINARY_DIR')
+        target.current_src_dir = self.var_to_str('MESON_PS_CMAKE_CURRENT_SOURCE_DIR')
         target.outputs = self._guess_files(target.outputs)
         target.depends = self._guess_files(target.depends)
         target.command = [self._guess_files(x) for x in target.command]
@@ -314,9 +390,7 @@ class CMakeTraceParser:
         # DOC: https://cmake.org/cmake/help/latest/command/set_property.html
         args = list(tline.args)
 
-        # We only care for TARGET properties
-        if args.pop(0) != 'TARGET':
-            return
+        scope = args.pop(0)
 
         append = False
         targets = []
@@ -331,7 +405,7 @@ class CMakeTraceParser:
             if curr == 'PROPERTY':
                 break
 
-            targets.append(curr)
+            targets += curr.split(';')
 
         if not args:
             return self._gen_exception('set_property', 'faild to parse argument list', tline)
@@ -341,11 +415,14 @@ class CMakeTraceParser:
             return
 
         identifier = args.pop(0)
-        value = ' '.join(args).split(';')
+        if self.trace_format == 'human':
+            value = ' '.join(args).split(';')
+        else:
+            value = [y for x in args for y in x.split(';')]
         if not value:
             return
 
-        for i in targets:
+        def do_target(tgt: str) -> None:
             if i not in self.targets:
                 return self._gen_exception('set_property', 'TARGET {} not found'.format(i), tline)
 
@@ -357,6 +434,33 @@ class CMakeTraceParser:
             else:
                 self.targets[i].properties[identifier] = value
 
+        def do_source(src: str) -> None:
+            if identifier != 'HEADER_FILE_ONLY' or not self._str_to_bool(value):
+                return
+
+            current_src_dir = self.var_to_str('MESON_PS_CMAKE_CURRENT_SOURCE_DIR')
+            if not current_src_dir:
+                mlog.warning(textwrap.dedent('''\
+                    CMake trace: set_property(SOURCE) called before the preload script was loaded.
+                    Unable to determine CMAKE_CURRENT_SOURCE_DIR. This can lead to build errors.
+                '''))
+                current_src_dir = '.'
+
+            cur_p = Path(current_src_dir)
+            src_p = Path(src)
+
+            if not src_p.is_absolute():
+                src_p = cur_p / src_p
+            self.explicit_headers.add(src_p)
+
+        if scope == 'TARGET':
+            for i in targets:
+                do_target(i)
+        elif scope == 'SOURCE':
+            files = self._guess_files(targets)
+            for i in files:
+                do_source(i)
+
     def _cmake_set_target_properties(self, tline: CMakeTraceLine) -> None:
         # DOC: https://cmake.org/cmake/help/latest/command/set_target_properties.html
         args = list(tline.args)
@@ -381,21 +485,27 @@ class CMakeTraceParser:
         # Neither of these is awesome for obvious reasons. I'm going to try
         # option 1 first and fall back to 2, as 1 requires less code and less
         # synchroniztion for cmake changes.
+        #
+        # With the JSON output format, introduced in CMake 3.17, spaces are
+        # handled properly and we don't have to do either options
 
         arglist = []  # type: T.List[T.Tuple[str, T.List[str]]]
-        name = args.pop(0)
-        values = []
-        prop_regex = re.compile(r'^[A-Z_]+$')
-        for a in args:
-            if prop_regex.match(a):
-                if values:
-                    arglist.append((name, ' '.join(values).split(';')))
-                name = a
-                values = []
-            else:
-                values.append(a)
-        if values:
-            arglist.append((name, ' '.join(values).split(';')))
+        if self.trace_format == 'human':
+            name = args.pop(0)
+            values = []
+            prop_regex = re.compile(r'^[A-Z_]+$')
+            for a in args:
+                if prop_regex.match(a):
+                    if values:
+                        arglist.append((name, ' '.join(values).split(';')))
+                    name = a
+                    values = []
+                else:
+                    values.append(a)
+            if values:
+                arglist.append((name, ' '.join(values).split(';')))
+        else:
+            arglist = [(x[0], x[1].split(';')) for x in zip(args[::2], args[1::2])]
 
         for name, value in arglist:
             for i in targets:
@@ -415,7 +525,8 @@ class CMakeTraceParser:
         if not target:
             return self._gen_exception('add_dependencies', 'target not found', tline)
 
-        target.depends += args[1:]
+        for i in args[1:]:
+            target.depends += i.split(';')
 
     def _cmake_target_compile_definitions(self, tline: CMakeTraceLine) -> None:
         # DOC: https://cmake.org/cmake/help/latest/command/target_compile_definitions.html
@@ -481,7 +592,19 @@ class CMakeTraceParser:
 
             self.targets[target].properties[i[0]] += i[1]
 
-    def _lex_trace(self, trace):
+    def _meson_ps_execute_delayed_calls(self, tline: CMakeTraceLine) -> None:
+        for l in self.stored_commands:
+            fn = self.functions.get(l.func, None)
+            if(fn):
+                fn(l)
+
+        # clear the stored commands
+        self.stored_commands = []
+
+    def _meson_ps_reload_vars(self, tline: CMakeTraceLine) -> None:
+        self.delayed_commands = self.get_cmake_var('MESON_PS_DELAYED_CALLS')
+
+    def _lex_trace_human(self, trace):
         # The trace format is: '<file>(<line>):  <func>(<args -- can contain \n> )\n'
         reg_tline = re.compile(r'\s*(.*\.(cmake|txt))\(([0-9]+)\):\s*(\w+)\(([\s\S]*?) ?\)\s*\n', re.MULTILINE)
         reg_other = re.compile(r'[^\n]*\n')
@@ -509,33 +632,68 @@ class CMakeTraceParser:
 
             yield CMakeTraceLine(file, line, func, args)
 
+    def _lex_trace_json(self, trace: str):
+        lines = trace.splitlines(keepends=False)
+        lines.pop(0)  # The first line is the version
+        for i in lines:
+            data = json.loads(i)
+            args = data['args']
+            args = [parse_generator_expressions(x) for x in args]
+            yield CMakeTraceLine(data['file'], data['line'], data['cmd'], args)
+
+    def _flatten_args(self, args: T.List[str]) -> T.List[str]:
+        # Split lists in arguments
+        res = []  # type: T.List[str]
+        for i in args:
+            res += i.split(';')
+        return res
+
     def _guess_files(self, broken_list: T.List[str]) -> T.List[str]:
-        #Try joining file paths that contain spaces
+        # Nothing has to be done for newer formats
+        if self.trace_format != 'human':
+            return broken_list
+
+        # Try joining file paths that contain spaces
 
         reg_start = re.compile(r'^([A-Za-z]:)?/.*/[^./]+$')
         reg_end = re.compile(r'^.*\.[a-zA-Z]+$')
 
         fixed_list = []  # type: T.List[str]
         curr_str = None  # type: T.Optional[str]
+        path_found = False # type: bool
 
         for i in broken_list:
             if curr_str is None:
                 curr_str = i
+                path_found = False
             elif os.path.isfile(curr_str):
                 # Abort concatenation if curr_str is an existing file
                 fixed_list += [curr_str]
                 curr_str = i
+                path_found = False
             elif not reg_start.match(curr_str):
                 # Abort concatenation if curr_str no longer matches the regex
                 fixed_list += [curr_str]
                 curr_str = i
-            elif reg_end.match(i) or os.path.exists('{} {}'.format(curr_str, i)):
+                path_found = False
+            elif reg_end.match(i):
                 # File detected
                 curr_str = '{} {}'.format(curr_str, i)
                 fixed_list += [curr_str]
                 curr_str = None
+                path_found = False
+            elif os.path.exists('{} {}'.format(curr_str, i)):
+                # Path detected
+                curr_str = '{} {}'.format(curr_str, i)
+                path_found = True
+            elif path_found:
+                # Add path to fixed_list after ensuring the whole path is in curr_str
+                fixed_list += [curr_str]
+                curr_str = i
+                path_found = False
             else:
                 curr_str = '{} {}'.format(curr_str, i)
+                path_found = False
 
         if curr_str:
             fixed_list += [curr_str]
