@@ -14,9 +14,14 @@
 
 import contextlib, os.path, re, tempfile
 import collections.abc
+import itertools
 import typing as T
+from functools import lru_cache
 
-from ..linkers import StaticLinker, GnuLikeDynamicLinkerMixin, SolarisDynamicLinker
+from ..linkers import (
+    GnuLikeDynamicLinkerMixin, LinkerEnvVarsMixin, SolarisDynamicLinker,
+    StaticLinker,
+)
 from .. import coredata
 from .. import mlog
 from .. import mesonlib
@@ -25,7 +30,7 @@ from ..mesonlib import (
     Popen_safe, split_args
 )
 from ..envconfig import (
-    Properties,
+    Properties, get_env_var
 )
 
 if T.TYPE_CHECKING:
@@ -34,6 +39,8 @@ if T.TYPE_CHECKING:
     from ..environment import Environment
     from ..linkers import DynamicLinker  # noqa: F401
 
+    CompilerType = T.TypeVar('CompilerType', bound=Compiler)
+
 """This file contains the data files of all compilers Meson knows
 about. To support a new compiler, add its information below.
 Also add corresponding autodetection code in environment.py."""
@@ -72,9 +79,12 @@ clink_suffixes = ()
 for _l in clink_langs + ('vala',):
     clink_suffixes += lang_suffixes[_l]
 clink_suffixes += ('h', 'll', 's')
+all_suffixes = set(itertools.chain(*lang_suffixes.values(), clink_suffixes))
 
 # Languages that should use LDFLAGS arguments when linking.
-languages_using_ldflags = ('objcpp', 'cpp', 'objc', 'c', 'fortran', 'd', 'cuda')
+languages_using_ldflags = {'objcpp', 'cpp', 'objc', 'c', 'fortran', 'd', 'cuda'}
+# Languages that should use CPPFLAGS arguments when linking.
+languages_using_cppflags = {'c', 'cpp', 'objc', 'objcpp'}
 soregex = re.compile(r'.*\.so(\.[0-9]+)?(\.[0-9]+)?(\.[0-9]+)?$')
 
 # Environment variables that each lang uses.
@@ -144,6 +154,13 @@ def is_library(fname):
     suffix = fname.split('.')[-1]
     return suffix in lib_suffixes
 
+def is_known_suffix(fname):
+    if hasattr(fname, 'fname'):
+        fname = fname.fname
+    suffix = fname.split('.')[-1]
+
+    return suffix in all_suffixes
+
 cuda_buildtype_args = {'plain': [],
                        'debug': [],
                        'debugoptimized': [],
@@ -445,6 +462,9 @@ class CompilerArgs(collections.abc.MutableSequence):
                  iterable: T.Optional[T.Iterable[str]] = None):
         self.compiler = compiler
         self.__container = list(iterable) if iterable is not None else []  # type: T.List[str]
+        self.__seen_args = set()
+        for arg in self.__container:
+            self.__seen_args.add(arg)
 
     @T.overload                                # noqa: F811
     def __getitem__(self, index: int) -> str:  # noqa: F811
@@ -467,20 +487,27 @@ class CompilerArgs(collections.abc.MutableSequence):
 
     def __setitem__(self, index, value) -> None:  # noqa: F811
         self.__container[index] = value
+        for v in value:
+            self.__seen_args.add(v)
 
     def __delitem__(self, index: T.Union[int, slice]) -> None:
+        value = self.__container[index]
         del self.__container[index]
+        if value in self.__seen_args and value in self.__container: # this is also honoring that you can have duplicated entries
+          self.__seen_args.remove(value)
 
     def __len__(self) -> int:
         return len(self.__container)
 
     def insert(self, index: int, value: str) -> None:
         self.__container.insert(index, value)
+        self.__seen_args.add(value)
 
     def copy(self) -> 'CompilerArgs':
         return CompilerArgs(self.compiler, self.__container.copy())
 
     @classmethod
+    @lru_cache(maxsize=None)
     def _can_dedup(cls, arg):
         '''
         Returns whether the argument can be safely de-duped. This is dependent
@@ -514,6 +541,15 @@ class CompilerArgs(collections.abc.MutableSequence):
         # both of which are invalid.
         if arg in cls.dedup2_prefixes:
             return 0
+        if arg.startswith('-L='):
+            # DMD and LDC proxy all linker arguments using -L=; in conjunction
+            # with ld64 on macOS this can lead to command line arguments such
+            # as: `-L=-compatibility_version -L=0 -L=current_version -L=0`.
+            # These cannot be combined, ld64 insists they must be passed with
+            # spaces and quoting does not work. if we deduplicate these then
+            # one of the -L=0 arguments will be removed and the version
+            # argument will consume the next argument instead.
+            return 0
         if arg in cls.dedup2_args or \
            arg.startswith(cls.dedup2_prefixes) or \
            arg.endswith(cls.dedup2_suffixes):
@@ -526,11 +562,15 @@ class CompilerArgs(collections.abc.MutableSequence):
         return 0
 
     @classmethod
+    @lru_cache(maxsize=None)
     def _should_prepend(cls, arg):
         if arg.startswith(cls.prepend_prefixes):
             return True
         return False
 
+    def need_to_split_linker_args(self):
+        return isinstance(self.compiler, Compiler) and self.compiler.get_language() == 'd'
+
     def to_native(self, copy: bool = False) -> T.List[str]:
         # Check if we need to add --start/end-group for circular dependencies
         # between static libraries, and for recursively searching for symbols
@@ -540,6 +580,10 @@ class CompilerArgs(collections.abc.MutableSequence):
             new = self.copy()
         else:
             new = self
+        # To proxy these arguments with D you need to split the
+        # arguments, thus you get `-L=-soname -L=lib.so` we don't
+        # want to put the lib in a link -roup
+        split_linker_args = self.need_to_split_linker_args()
         # This covers all ld.bfd, ld.gold, ld.gold, and xild on Linux, which
         # all act like (or are) gnu ld
         # TODO: this could probably be added to the DynamicLinker instead
@@ -548,7 +592,14 @@ class CompilerArgs(collections.abc.MutableSequence):
                 isinstance(self.compiler.linker, (GnuLikeDynamicLinkerMixin, SolarisDynamicLinker))):
             group_start = -1
             group_end = -1
+            is_soname = False
             for i, each in enumerate(new):
+                if is_soname:
+                    is_soname = False
+                    continue
+                elif split_linker_args and '-soname' in each:
+                    is_soname = True
+                    continue
                 if not each.startswith(('-Wl,-l', '-l')) and not each.endswith('.a') and \
                    not soregex.match(each):
                     continue
@@ -588,6 +639,7 @@ class CompilerArgs(collections.abc.MutableSequence):
             self.append(arg)
         else:
             self.__container.append(arg)
+            self.__seen_args.add(arg)
 
     def extend_direct(self, iterable: T.Iterable[str]) -> None:
         '''
@@ -619,6 +671,7 @@ class CompilerArgs(collections.abc.MutableSequence):
         Add two CompilerArgs while taking into account overriding of arguments
         and while preserving the order of arguments as much as possible
         '''
+        this_round_added = set() # a dict that contains a value, when the value was added this round
         pre = []   # type: T.List[str]
         post = []  # type: T.List[str]
         if not isinstance(args, collections.abc.Iterable):
@@ -630,20 +683,25 @@ class CompilerArgs(collections.abc.MutableSequence):
             dedup = self._can_dedup(arg)
             if dedup == 1:
                 # Argument already exists and adding a new instance is useless
-                if arg in self or arg in pre or arg in post:
+                if arg in self.__seen_args or arg in pre or arg in post:
                     continue
+            should_prepend = self._should_prepend(arg)
             if dedup == 2:
                 # Remove all previous occurrences of the arg and add it anew
-                if arg in self:
+                if arg in self.__seen_args and arg not in this_round_added: #if __seen_args contains arg as well as this_round_added, then its not yet part in self.
                     self.remove(arg)
-                if arg in pre:
-                    pre.remove(arg)
-                if arg in post:
-                    post.remove(arg)
-            if self._should_prepend(arg):
+                if should_prepend:
+                    if arg in pre:
+                        pre.remove(arg)
+                else:
+                    if arg in post:
+                        post.remove(arg)
+            if should_prepend:
                 pre.append(arg)
             else:
                 post.append(arg)
+            self.__seen_args.add(arg)
+            this_round_added.add(arg)
         # Insert at the beginning
         self[:0] = pre
         # Append to the end
@@ -712,6 +770,7 @@ class Compiler:
         return repr_str.format(self.__class__.__name__, self.version,
                                ' '.join(self.exelist))
 
+    @lru_cache(maxsize=None)
     def can_compile(self, src) -> bool:
         if hasattr(src, 'fname'):
             src = src.fname
@@ -724,7 +783,13 @@ class Compiler:
         return self.id
 
     def get_linker_id(self) -> str:
-        return self.linker.id
+        # There is not guarantee that we have a dynamic linker instance, as
+        # some languages don't have separate linkers and compilers. In those
+        # cases return the compiler id
+        try:
+            return self.linker.id
+        except AttributeError:
+            return self.id
 
     def get_version_string(self) -> str:
         details = [self.id, self.version]
@@ -786,7 +851,7 @@ class Compiler:
         """
         return self.linker.get_accepts_rsp()
 
-    def get_linker_always_args(self):
+    def get_linker_always_args(self) -> T.List[str]:
         return self.linker.get_always_args()
 
     def get_linker_lib_prefix(self):
@@ -799,8 +864,10 @@ class Compiler:
         """
         return []
 
-    def get_linker_args_from_envvars(self) -> T.List[str]:
-        return self.linker.get_args_from_envvars()
+    def get_linker_args_from_envvars(self,
+                                     for_machine: MachineChoice,
+                                     is_cross: bool) -> T.List[str]:
+        return self.linker.get_args_from_envvars(for_machine, is_cross)
 
     def get_options(self) -> T.Dict[str, coredata.UserOption]:
         return {}
@@ -976,6 +1043,9 @@ class Compiler:
     def get_compile_debugfile_args(self, rel_obj, **kwargs):
         return []
 
+    def get_link_debugfile_name(self, targetfile: str) -> str:
+        return self.linker.get_debugfile_name(targetfile)
+
     def get_link_debugfile_args(self, targetfile: str) -> T.List[str]:
         return self.linker.get_debugfile_args(targetfile)
 
@@ -1059,7 +1129,25 @@ class Compiler:
         return self.linker.get_undefined_link_args()
 
     def remove_linkerlike_args(self, args):
-        return [x for x in args if not x.startswith('-Wl')]
+        rm_exact = ('-headerpad_max_install_names',)
+        rm_prefixes = ('-Wl,', '-L',)
+        rm_next = ('-L', '-framework',)
+        ret = []
+        iargs = iter(args)
+        for arg in iargs:
+            # Remove this argument
+            if arg in rm_exact:
+                continue
+            # If the argument starts with this, but is not *exactly* this
+            # f.ex., '-L' should match ['-Lfoo'] but not ['-L', 'foo']
+            if arg.startswith(rm_prefixes) and arg not in rm_prefixes:
+                continue
+            # Ignore this argument and the one after it
+            if arg in rm_next:
+                next(iargs)
+                continue
+            ret.append(arg)
+        return ret
 
     def get_lto_compile_args(self) -> T.List[str]:
         return []
@@ -1131,37 +1219,27 @@ def get_largefile_args(compiler):
     return []
 
 
-def get_args_from_envvars(lang: str, use_linker_args: bool) -> T.Tuple[T.List[str], T.List[str]]:
+def get_args_from_envvars(lang: str,
+                          for_machine: MachineChoice,
+                          is_cross: bool,
+                          use_linker_args: bool) -> T.Tuple[T.List[str], T.List[str]]:
     """
     Returns a tuple of (compile_flags, link_flags) for the specified language
     from the inherited environment
     """
-    def log_var(var, val: T.Optional[str]):
-        if val:
-            mlog.log('Appending {} from environment: {!r}'.format(var, val))
-        else:
-            mlog.debug('No {} in the environment, not changing global flags.'.format(var))
-
     if lang not in cflags_mapping:
         return [], []
 
     compile_flags = []  # type: T.List[str]
     link_flags = []     # type: T.List[str]
 
-    env_compile_flags = os.environ.get(cflags_mapping[lang])
-    log_var(cflags_mapping[lang], env_compile_flags)
+    env_compile_flags = get_env_var(for_machine, is_cross, cflags_mapping[lang])
     if env_compile_flags is not None:
         compile_flags += split_args(env_compile_flags)
 
     # Link flags (same for all languages)
     if lang in languages_using_ldflags:
-        # This is duplicated between the linkers, but I'm not sure how else
-        # to handle this
-        env_link_flags = split_args(os.environ.get('LDFLAGS', ''))
-    else:
-        env_link_flags = []
-    log_var('LDFLAGS', env_link_flags)
-    link_flags += env_link_flags
+        link_flags += LinkerEnvVarsMixin.get_args_from_envvars(for_machine, is_cross)
     if use_linker_args:
         # When the compiler is used as a wrapper around the linker (such as
         # with GCC and Clang), the compile flags can be needed while linking
@@ -1170,16 +1248,18 @@ def get_args_from_envvars(lang: str, use_linker_args: bool) -> T.Tuple[T.List[st
         link_flags = compile_flags + link_flags
 
     # Pre-processor flags for certain languages
-    if lang in {'c', 'cpp', 'objc', 'objcpp'}:
-        env_preproc_flags = os.environ.get('CPPFLAGS')
-        log_var('CPPFLAGS', env_preproc_flags)
+    if lang in languages_using_cppflags:
+        env_preproc_flags = get_env_var(for_machine, is_cross, 'CPPFLAGS')
         if env_preproc_flags is not None:
             compile_flags += split_args(env_preproc_flags)
 
     return compile_flags, link_flags
 
 
-def get_global_options(lang: str, comp: T.Type[Compiler],
+def get_global_options(lang: str,
+                       comp: T.Type[Compiler],
+                       for_machine: MachineChoice,
+                       is_cross: bool,
                        properties: Properties) -> T.Dict[str, coredata.UserOption]:
     """Retreive options that apply to all compilers for a given language."""
     description = 'Extra arguments passed to the {}'.format(lang)
@@ -1192,13 +1272,12 @@ def get_global_options(lang: str, comp: T.Type[Compiler],
             [], split_args=True, user_input=True, allow_dups=True),
     }
 
-    if properties.fallback:
-        # Get from env vars.
-        # XXX: True here is a hack
-        compile_args, link_args = get_args_from_envvars(lang, comp.INVOKES_LINKER)
-    else:
-        compile_args = []
-        link_args = []
+    # Get from env vars.
+    compile_args, link_args = get_args_from_envvars(
+        lang,
+        for_machine,
+        is_cross,
+        comp.INVOKES_LINKER)
 
     for k, o in opts.items():
         if k in properties:
