@@ -13,13 +13,13 @@
 # limitations under the License.
 
 import configparser, os, subprocess
-import typing
+import typing as T
 
 from . import mesonlib
-from .mesonlib import EnvironmentException, split_args
+from .mesonlib import EnvironmentException, MachineChoice, PerMachine, split_args
 from . import mlog
 
-_T = typing.TypeVar('_T')
+_T = T.TypeVar('_T')
 
 
 # These classes contains all the data pulled from configuration files (native
@@ -40,8 +40,11 @@ known_cpu_families = (
     'alpha',
     'arc',
     'arm',
+    'avr',
+    'c2000',
     'e2k',
     'ia64',
+    'm68k',
     'microblaze',
     'mips',
     'mips64',
@@ -52,9 +55,12 @@ known_cpu_families = (
     'riscv64',
     'rl78',
     'rx',
+    's390',
     's390x',
     'sparc',
     'sparc64',
+    'pic24',
+    'dspic',
     'wasm32',
     'wasm64',
     'x86',
@@ -70,6 +76,7 @@ CPU_FAMILES_64_BIT = [
     'mips64',
     'ppc64',
     'riscv64',
+    's390x',
     'sparc64',
     'wasm64',
     'x86_64',
@@ -77,7 +84,7 @@ CPU_FAMILES_64_BIT = [
 
 class MesonConfigFile:
     @classmethod
-    def from_config_parser(cls, parser: configparser.ConfigParser) -> typing.Dict[str, typing.Dict[str, typing.Dict[str, str]]]:
+    def from_config_parser(cls, parser: configparser.ConfigParser) -> T.Dict[str, T.Dict[str, T.Dict[str, str]]]:
         out = {}
         # This is a bit hackish at the moment.
         for s in parser.sections():
@@ -102,28 +109,48 @@ class MesonConfigFile:
             out[s] = section
         return out
 
-class HasEnvVarFallback:
+def get_env_var_pair(for_machine: MachineChoice,
+                     is_cross: bool,
+                     var_name: str) -> T.Tuple[T.Optional[str], T.Optional[str]]:
     """
-    A tiny class to indicate that this class contains data that can be
-    initialized from either a config file or environment file. The `fallback`
-    field says whether env vars should be used. Downstream logic (e.g. subclass
-    methods) can check it to decide what to do, since env vars are currently
-    lazily decoded.
-
-    Frankly, this is a pretty silly class at the moment. The hope is the way
-    that we deal with environment variables will become more structured, and
-    this can be starting point.
+    Returns the exact env var and the value.
     """
-    def __init__(self, fallback: bool = True):
-        self.fallback = fallback
+    candidates = PerMachine(
+        # The prefixed build version takes priority, but if we are native
+        # compiling we fall back on the unprefixed host version. This
+        # allows native builds to never need to worry about the 'BUILD_*'
+        # ones.
+        ([var_name + '_FOR_BUILD'] if is_cross else [var_name]),
+        # Always just the unprefixed host verions
+        [var_name]
+    )[for_machine]
+    for var in candidates:
+        value = os.environ.get(var)
+        if value is not None:
+            break
+    else:
+        formatted = ', '.join(['{!r}'.format(var) for var in candidates])
+        mlog.debug('None of {} are defined in the environment, not changing global flags.'.format(formatted))
+        return None
+    mlog.log('Using {!r} from environment with value: {!r}'.format(var, value))
+    return var, value
 
-class Properties(HasEnvVarFallback):
+def get_env_var(for_machine: MachineChoice,
+                is_cross: bool,
+                var_name: str) -> T.Tuple[T.Optional[str], T.Optional[str]]:
+    ret = get_env_var_pair(for_machine, is_cross, var_name)
+    if ret is None:
+        return None
+    else:
+        var, value = ret
+        return value
+
+class Properties:
     def __init__(
             self,
-            properties: typing.Optional[typing.Dict[str, typing.Union[str, typing.List[str]]]] = None,
-            fallback: bool = True):
-        super().__init__(fallback)
-        self.properties = properties or {}  # type: typing.Dict[str, typing.Union[str, typing.List[str]]]
+            properties: T.Optional[T.Dict[str, T.Union[str, T.List[str]]]] = None,
+    ):
+        self.properties = properties or {}  # type: T.Dict[str, T.Union[str, T.List[str]]]
 
     def has_stdlib(self, language: str) -> bool:
         return language + '_stdlib' in self.properties
@@ -131,30 +158,36 @@ class Properties(HasEnvVarFallback):
     # Some of get_stdlib, get_root, get_sys_root are wider than is actually
     # true, but without heterogenious dict annotations it's not practical to
     # narrow them
-    def get_stdlib(self, language: str) -> typing.Union[str, typing.List[str]]:
+    def get_stdlib(self, language: str) -> T.Union[str, T.List[str]]:
         return self.properties[language + '_stdlib']
 
-    def get_root(self) -> typing.Optional[typing.Union[str, typing.List[str]]]:
+    def get_root(self) -> T.Optional[T.Union[str, T.List[str]]]:
         return self.properties.get('root', None)
 
-    def get_sys_root(self) -> typing.Optional[typing.Union[str, typing.List[str]]]:
+    def get_sys_root(self) -> T.Optional[T.Union[str, T.List[str]]]:
         return self.properties.get('sys_root', None)
 
-    def __eq__(self, other: typing.Any) -> 'typing.Union[bool, NotImplemented]':
+    def get_pkg_config_libdir(self) -> T.Optional[T.List[str]]:
+        p = self.properties.get('pkg_config_libdir', None)
+        if p is None:
+            return p
+        return mesonlib.listify(p)
+
+    def __eq__(self, other: T.Any) -> 'T.Union[bool, NotImplemented]':
         if isinstance(other, type(self)):
             return self.properties == other.properties
         return NotImplemented
 
     # TODO consider removing so Properties is less freeform
-    def __getitem__(self, key: str) -> typing.Any:
+    def __getitem__(self, key: str) -> T.Any:
         return self.properties[key]
 
     # TODO consider removing so Properties is less freeform
-    def __contains__(self, item: typing.Any) -> bool:
+    def __contains__(self, item: T.Any) -> bool:
         return item in self.properties
 
     # TODO consider removing, for same reasons as above
-    def get(self, key: str, default: typing.Any = None) -> typing.Any:
+    def get(self, key: str, default: T.Any = None) -> T.Any:
         return self.properties.get(key, default)
 
 class MachineInfo:
@@ -165,7 +198,7 @@ class MachineInfo:
         self.endian = endian
         self.is_64_bit = cpu_family in CPU_FAMILES_64_BIT  # type: bool
 
-    def __eq__(self, other: typing.Any) -> 'typing.Union[bool, NotImplemented]':
+    def __eq__(self, other: T.Any) -> 'T.Union[bool, NotImplemented]':
         if self.__class__ is not other.__class__:
             return NotImplemented
         return \
@@ -174,7 +207,7 @@ class MachineInfo:
             self.cpu == other.cpu and \
             self.endian == other.endian
 
-    def __ne__(self, other: typing.Any) -> 'typing.Union[bool, NotImplemented]':
+    def __ne__(self, other: T.Any) -> 'T.Union[bool, NotImplemented]':
         if self.__class__ is not other.__class__:
             return NotImplemented
         return not self.__eq__(other)
@@ -183,7 +216,7 @@ class MachineInfo:
         return '<MachineInfo: {} {} ({})>'.format(self.system, self.cpu_family, self.cpu)
 
     @classmethod
-    def from_literal(cls, literal: typing.Dict[str, str]) -> 'MachineInfo':
+    def from_literal(cls, literal: T.Dict[str, str]) -> 'MachineInfo':
         minimum_literal = {'cpu', 'cpu_family', 'endian', 'system'}
         if set(literal) < minimum_literal:
             raise EnvironmentException(
@@ -278,13 +311,12 @@ class MachineInfo:
     def libdir_layout_is_win(self) -> bool:
         return self.is_windows() or self.is_cygwin()
 
-class BinaryTable(HasEnvVarFallback):
+class BinaryTable:
     def __init__(
             self,
-            binaries: typing.Optional[typing.Dict[str, typing.Union[str, typing.List[str]]]] = None,
-            fallback: bool = True):
-        super().__init__(fallback)
-        self.binaries = binaries or {}  # type: typing.Dict[str, typing.Union[str, typing.List[str]]]
+            binaries: T.Optional[T.Dict[str, T.Union[str, T.List[str]]]] = None,
+    ):
+        self.binaries = binaries or {}  # type: T.Dict[str, T.Union[str, T.List[str]]]
         for name, command in self.binaries.items():
             if not isinstance(command, (list, str)):
                 # TODO generalize message
@@ -305,20 +337,37 @@ class BinaryTable(HasEnvVarFallback):
         'rust': 'RUSTC',
         'vala': 'VALAC',
 
+        # Linkers
+        'c_ld': 'CC_LD',
+        'cpp_ld': 'CXX_LD',
+        'd_ld': 'DC_LD',
+        'fortran_ld': 'FC_LD',
+        'objc_ld': 'OBJC_LD',
+        'objcpp_ld': 'OBJCXX_LD',
+        'rust_ld': 'RUSTC_LD',
+
         # Binutils
         'strip': 'STRIP',
         'ar': 'AR',
         'windres': 'WINDRES',
-        'ld': 'LD',
 
         # Other tools
         'cmake': 'CMAKE',
         'qmake': 'QMAKE',
         'pkgconfig': 'PKG_CONFIG',
-    }  # type: typing.Dict[str, str]
+    }  # type: T.Dict[str, str]
+
+    # Deprecated environment variables mapped from the new variable to the old one
+    # Deprecated in 0.54.0
+    DEPRECATION_MAP = {
+        'DC_LD': 'D_LD',
+        'FC_LD': 'F_LD',
+        'RUSTC_LD': 'RUST_LD',
+        'OBJCXX_LD': 'OBJCPP_LD',
+    }  # type: T.Dict[str, str]
 
     @staticmethod
-    def detect_ccache() -> typing.List[str]:
+    def detect_ccache() -> T.List[str]:
         try:
             subprocess.check_call(['ccache', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
         except (OSError, subprocess.CalledProcessError):
@@ -326,14 +375,7 @@ class BinaryTable(HasEnvVarFallback):
         return ['ccache']
 
     @classmethod
-    def _warn_about_lang_pointing_to_cross(cls, compiler_exe: str, evar: str) -> None:
-        evar_str = os.environ.get(evar, 'WHO_WOULD_CALL_THEIR_COMPILER_WITH_THIS_NAME')
-        if evar_str == compiler_exe:
-            mlog.warning('''Env var %s seems to point to the cross compiler.
-This is probably wrong, it should always point to the native compiler.''' % evar)
-
-    @classmethod
-    def parse_entry(cls, entry: typing.Union[str, typing.List[str]]) -> typing.Tuple[typing.List[str], typing.List[str]]:
+    def parse_entry(cls, entry: T.Union[str, T.List[str]]) -> T.Tuple[T.List[str], T.List[str]]:
         compiler = mesonlib.stringlistify(entry)
         # Ensure ccache exists and remove it if it doesn't
         if compiler[0] == 'ccache':
@@ -344,31 +386,42 @@ This is probably wrong, it should always point to the native compiler.''' % evar
         # Return value has to be a list of compiler 'choices'
         return compiler, ccache
 
-    def lookup_entry(self, name: str) -> typing.Optional[typing.List[str]]:
-        """Lookup binaryk
+    def lookup_entry(self,
+                     for_machine: MachineChoice,
+                     is_cross: bool,
+                     name: str) -> T.Optional[T.List[str]]:
+        """Lookup binary in cross/native file and fallback to environment.
 
         Returns command with args as list if found, Returns `None` if nothing is
         found.
-
-        First tries looking in explicit map, then tries environment variable.
         """
         # Try explicit map, don't fall back on env var
-        command = self.binaries.get(name)
-        if command is not None:
-            command = mesonlib.stringlistify(command)
-            # Relies on there being no "" env var
-            evar = self.evarMap.get(name, "")
-            self._warn_about_lang_pointing_to_cross(command[0], evar)
-        elif self.fallback:
-            # Relies on there being no "" env var
-            evar = self.evarMap.get(name, "")
-            command = os.environ.get(evar)
-            if command is not None:
-                command = split_args(command)
+        # Try explict map, then env vars
+        for _ in [()]: # a trick to get `break`
+            raw_command = self.binaries.get(name)
+            if raw_command is not None:
+                command = mesonlib.stringlistify(raw_command)
+                break # found
+            evar = self.evarMap.get(name)
+            if evar is not None:
+                raw_command = get_env_var(for_machine, is_cross, evar)
+                if raw_command is None:
+                    deprecated = self.DEPRECATION_MAP.get(evar)
+                    if deprecated is not None:
+                        raw_command = get_env_var(for_machine, is_cross, deprecated)
+                        if raw_command is not None:
+                            mlog.deprecation(
+                                'The', deprecated, 'environment variable is deprecated in favor of',
+                                evar, once=True)
+                if raw_command is not None:
+                    command = split_args(raw_command)
+                    break # found
+            command = None
+
 
         # Do not return empty or blank string entries
         if command is not None and (len(command) == 0 or len(command[0].strip()) == 0):
-            return None
+            command = None
         return command
 
 class Directories:
@@ -377,13 +430,13 @@ class Directories:
     builds.
     """
 
-    def __init__(self, bindir: typing.Optional[str] = None, datadir: typing.Optional[str] = None,
-                 includedir: typing.Optional[str] = None, infodir: typing.Optional[str] = None,
-                 libdir: typing.Optional[str] = None, libexecdir: typing.Optional[str] = None,
-                 localedir: typing.Optional[str] = None, localstatedir: typing.Optional[str] = None,
-                 mandir: typing.Optional[str] = None, prefix: typing.Optional[str] = None,
-                 sbindir: typing.Optional[str] = None, sharedstatedir: typing.Optional[str] = None,
-                 sysconfdir: typing.Optional[str] = None):
+    def __init__(self, bindir: T.Optional[str] = None, datadir: T.Optional[str] = None,
+                 includedir: T.Optional[str] = None, infodir: T.Optional[str] = None,
+                 libdir: T.Optional[str] = None, libexecdir: T.Optional[str] = None,
+                 localedir: T.Optional[str] = None, localstatedir: T.Optional[str] = None,
+                 mandir: T.Optional[str] = None, prefix: T.Optional[str] = None,
+                 sbindir: T.Optional[str] = None, sharedstatedir: T.Optional[str] = None,
+                 sysconfdir: T.Optional[str] = None):
         self.bindir = bindir
         self.datadir = datadir
         self.includedir = includedir
@@ -401,12 +454,12 @@ class Directories:
     def __contains__(self, key: str) -> bool:
         return hasattr(self, key)
 
-    def __getitem__(self, key: str) -> typing.Optional[str]:
+    def __getitem__(self, key: str) -> T.Optional[str]:
         # Mypy can't figure out what to do with getattr here, so we'll case for it
-        return typing.cast(typing.Optional[str], getattr(self, key))
+        return T.cast(T.Optional[str], getattr(self, key))
 
-    def __setitem__(self, key: str, value: typing.Optional[str]) -> None:
+    def __setitem__(self, key: str, value: T.Optional[str]) -> None:
         setattr(self, key, value)
 
-    def __iter__(self) -> typing.Iterator[typing.Tuple[str, str]]:
+    def __iter__(self) -> T.Iterator[T.Tuple[str, str]]:
         return iter(self.__dict__.items())
